package com.webchat.client.chat.config.interceptor;

import com.webchat.common.config.annotation.SafeClick;
import com.webchat.common.constants.WebConstant;
import com.webchat.common.exception.AuthException;
import com.webchat.common.exception.BusinessException;
import com.webchat.common.helper.SessionHelper;
import com.webchat.common.service.RedisService;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.method.HandlerMethod;
import org.springframework.web.servlet.HandlerInterceptor;

import java.lang.reflect.Method;

/**
 * 安全点击自定义注解实现，拦截器实现
 *
 * 实现原理：基于redis分布式实现限制当前登录用于在一定时间内只能触发一次特定事件，超过限制则报错提醒
 */
@Slf4j
public class SafeClickInterceptor implements HandlerInterceptor {


    @Autowired
    private RedisService redisService;

    public boolean preHandle(HttpServletRequest request, HttpServletResponse response, Object handler) {

        // 检查 handler 是否为 HandlerMethod 类型
        if (!(handler instanceof HandlerMethod)) {
            // 不是 HandlerMethod 类型，可能是静态资源请求，直接放行
            return true;
        }
        HandlerMethod handlerMethod = (HandlerMethod) handler;
        Method method = handlerMethod.getMethod();
        // 获取当前被拦截方法是否由SafeClick
        SafeClick safeClick = method.getAnnotation(SafeClick.class);
        if (safeClick == null) {
            // 没有添加@SafeClick，则不需要校验安全点击
            return true;
        }
        // 获取操作人(当前登录人) ====>
        // 当前请求在主线程内（tomcat线程池分配主线程）,所有可以从ThreadLocal中直接获取
        String userId = SessionHelper.getCurrentUserId();
        if (StringUtils.isBlank(userId)) {
            // 当前用户未登录(实际上合理不会触发未登录，因为我们在filter中刚完整了SSO登录认证)
            // PS：但是为保证逻辑的严谨性，这里还是做了判空
            throw new AuthException("未登录");
        }
        // 获取用户安全点击事件类型
        String action = safeClick.event().name();
        // 允许多久内操作一次
        Long time = safeClick.time();
        // 构造redis分布式锁redis key【规则：事件名称-用户id】
        String redisKey = action.concat("-").concat(userId);
        boolean installLock = redisService.installLockForMS(redisKey, WebConstant.CACHE_NONE, time);
        if (installLock) {
            // 分布式加锁成功，说明当前用户在请求可以允许放行，反之则触发限流
            return true;
        }
        log.error("<<< {} >>>, userId: {}, action:{}", safeClick.message(), userId, action);
        // 触发限流，抛异常，异常提示来自用户注解自定义异常描述
        throw new BusinessException(safeClick.message());
    }
}
