package com.webchat.ugc.service.redpacket;

import com.webchat.common.constants.RedPacketConstants;
import com.webchat.common.enums.RedisKeyEnum;
import com.webchat.common.exception.BusinessException;
import com.webchat.domain.vo.response.RedPacketBaseVO;
import org.apache.commons.lang3.StringUtils;
import org.springframework.data.redis.core.script.RedisScript;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;


/**
 * 方案二：自定义lua脚本由redis底层命令单线程执行来保证并发安全
 */
@Service
public class OpenRedPacketWithLuaService extends AbstractOpenRedPacketService {


    private static final String OPEN_LUA_SCRIPT =
            "-- 1. 接收execute参数" +
            "local amountKey = KEYS[1] -- 红包余额缓存KEY" +
            "local gradCountKey = KEYS[2] -- 红包剩余可以拆分份额缓存KEY" +
            "local openUsersKey = KEYS[3] -- 红包余额缓存KEY" +
            "local redPacketId = ARGV[1]  -- 红包ID" +
            "local userId = ARGV[2]       -- 拆包ID" +
            "-- 2. 校验是否由拆包资格" +
            "local balanceAmount = tonumber(redis.call('HGET', amountKey, redPacketId)) or 0 " +
            "if balanceAmount <= 0 " +
            "   return '' " +
            "local gradCount = tonumber(redis.call('HGET', gradCountKey, redPacketId)) or 0 " +
            "if gradCount <= 0 " +
            "   return '' " +
            "-- 3. 当前剩余可拆分份额为1分，直接返回剩余金额" +
            "if gradCount == 1 then "+
            "   redis.call('HSET', amountKey, redPacketId, 0) "+
            "   redis.call('HSET', gradCountKey, redPacketId, 0) "+
            "   redis.call('SADD', openUsersKey, userId) "+
            "   return tostring(balanceAmount) "+
            "-- 4. 基于二倍均值红包算法，为当前拆包线程随机计算红包金额"+
            "local minAmount = 1 " +
            "local avgAmount = math.floor(balanceAmount / gradCount) " +
            "local avgAmount2 = avgAmount * 2 " +
            "local openAmount = math.random(minAmount, avgAmount2 + 1) " +
            "   redis.call('HSET', amountKey, redPacketId, balanceAmount - openAmount ) "+
            "   redis.call('HSET', gradCountKey, redPacketId, gradCount - 1) "+
            "   redis.call('SADD', openUsersKey, userId) "+
            "   return tostring(openAmount) "
            ;

    @Override
    public int openRedPacket(RedPacketBaseVO redPacket, String userId) {
        Long redPacketId = redPacket.getId();

        List<String> keys = new ArrayList<>();
        // KEY1：红包剩余可拆分金额
        keys.add(RedisKeyEnum.RED_PACKET_BALANCE_COUNT.getKey());
        // KEY2：红包剩余可拆分份额
        keys.add(RedisKeyEnum.RED_PACKET_GRAD_COUNT.getKey());
        // KEY3：红包对应拆分用户名单缓存
        keys.add(RedisKeyEnum.RED_PACKET_OPEN_USERS.getKey(String.valueOf(redPacketId)));

        String amount = redisService.executeScript(
                RedisScript.of(OPEN_LUA_SCRIPT, String.class),
                keys,
                redPacketId, userId);
        if (StringUtils.isBlank(amount)) {
            super.updateRedPacketStatus(redPacketId, RedPacketConstants.RedPacketStatus.END);
            throw new BusinessException("来晚了");
        }
       return Integer.valueOf(amount);
    }
}
